package engineconfig_test

import (
	"io/fs"
	"testing"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/config"
	engineconfig "git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/config/engine"
	shardconfig "git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/config/engine/shard"
	blobovniczaconfig "git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/config/engine/shard/blobstor/blobovnicza"
	fstreeconfig "git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/config/engine/shard/blobstor/fstree"
	gcconfig "git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/config/engine/shard/gc"
	piloramaconfig "git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/config/engine/shard/pilorama"
	configtest "git.frostfs.info/TrueCloudLab/frostfs-node/cmd/frostfs-node/config/test"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/shard/mode"
	"github.com/stretchr/testify/require"
)

func TestEngineSection(t *testing.T) {
	t.Run("defaults", func(t *testing.T) {
		empty := configtest.EmptyConfig()

		require.ErrorIs(t,
			engineconfig.IterateShards(empty, true, nil),
			engineconfig.ErrNoShardConfigured)

		handlerCalled := false

		require.NoError(t,
			engineconfig.IterateShards(empty, false, func(_ *shardconfig.Config) error {
				handlerCalled = true
				return nil
			}))

		require.False(t, handlerCalled)

		require.EqualValues(t, 0, engineconfig.ShardErrorThreshold(empty))
		require.EqualValues(t, engineconfig.ShardPoolSizeDefault, engineconfig.ShardPoolSize(empty))
		require.EqualValues(t, mode.ReadWrite, shardconfig.From(empty).Mode())
		require.EqualValues(t, engineconfig.RebuildWorkersCountDefault, engineconfig.EngineRebuildWorkersCount(empty))
	})

	const path = "../../../../config/example/node"

	fileConfigTest := func(c *config.Config) {
		num := 0

		require.EqualValues(t, 100, engineconfig.ShardErrorThreshold(c))
		require.EqualValues(t, 15, engineconfig.ShardPoolSize(c))
		require.EqualValues(t, uint32(1000), engineconfig.EngineRebuildWorkersCount(c))

		err := engineconfig.IterateShards(c, true, func(sc *shardconfig.Config) error {
			defer func() {
				num++
			}()

			wc := sc.WriteCache()
			meta := sc.Metabase()
			blob := sc.BlobStor()
			ss := blob.Storages()
			pl := sc.Pilorama()
			gc := sc.GC()

			switch num {
			case 0:
				require.Equal(t, "tmp/0/blob/pilorama.db", pl.Path())
				require.Equal(t, fs.FileMode(piloramaconfig.PermDefault), pl.Perm())
				require.False(t, pl.NoSync())
				require.Equal(t, pl.MaxBatchDelay(), 10*time.Millisecond)
				require.Equal(t, pl.MaxBatchSize(), 200)

				require.Equal(t, false, wc.Enabled())
				require.Equal(t, true, wc.NoSync())

				require.Equal(t, "tmp/0/cache", wc.Path())
				require.EqualValues(t, 16384, wc.SmallObjectSize())
				require.EqualValues(t, 134217728, wc.MaxObjectSize())
				require.EqualValues(t, 30, wc.WorkerCount())
				require.EqualValues(t, 3221225472, wc.SizeLimit())

				require.Equal(t, "tmp/0/meta", meta.Path())
				require.Equal(t, fs.FileMode(0o644), meta.BoltDB().Perm())
				require.Equal(t, 100, meta.BoltDB().MaxBatchSize())
				require.Equal(t, 10*time.Millisecond, meta.BoltDB().MaxBatchDelay())

				require.Equal(t, true, sc.Compress())
				require.Equal(t, []string{"audio/*", "video/*"}, sc.UncompressableContentTypes())
				require.Equal(t, true, sc.EstimateCompressibility())
				require.Equal(t, float64(0.7), sc.EstimateCompressibilityThreshold())
				require.EqualValues(t, 102400, sc.SmallSizeLimit())

				require.Equal(t, 2, len(ss))
				blz := blobovniczaconfig.From((*config.Config)(ss[0]))
				require.Equal(t, "tmp/0/blob/blobovnicza", ss[0].Path())
				require.EqualValues(t, 0o644, blz.BoltDB().Perm())
				require.EqualValues(t, 4194304, blz.Size())
				require.EqualValues(t, 1, blz.ShallowDepth())
				require.EqualValues(t, 4, blz.ShallowWidth())
				require.EqualValues(t, 50, blz.OpenedCacheSize())
				require.EqualValues(t, 10, blz.LeafWidth())
				require.EqualValues(t, 10, blz.InitWorkerCount())
				require.EqualValues(t, true, blz.InitInAdvance())
				require.EqualValues(t, 30*time.Second, blz.RebuildDropTimeout())

				require.Equal(t, "tmp/0/blob", ss[1].Path())
				require.EqualValues(t, 0o644, ss[1].Perm())

				fst := fstreeconfig.From((*config.Config)(ss[1]))
				require.EqualValues(t, 5, fst.Depth())
				require.Equal(t, false, fst.NoSync())

				require.EqualValues(t, 150, gc.RemoverBatchSize())
				require.Equal(t, 2*time.Minute, gc.RemoverSleepInterval())
				require.Equal(t, 1500, gc.ExpiredCollectorBatchSize())
				require.Equal(t, 15, gc.ExpiredCollectorWorkerCount())

				require.Equal(t, false, sc.RefillMetabase())
				require.Equal(t, mode.ReadOnly, sc.Mode())
			case 1:
				require.Equal(t, "tmp/1/blob/pilorama.db", pl.Path())
				require.Equal(t, fs.FileMode(0o644), pl.Perm())
				require.True(t, pl.NoSync())
				require.Equal(t, 5*time.Millisecond, pl.MaxBatchDelay())
				require.Equal(t, 100, pl.MaxBatchSize())

				require.Equal(t, true, wc.Enabled())
				require.Equal(t, false, wc.NoSync())

				require.Equal(t, "tmp/1/cache", wc.Path())
				require.EqualValues(t, 16384, wc.SmallObjectSize())
				require.EqualValues(t, 134217728, wc.MaxObjectSize())
				require.EqualValues(t, 30, wc.WorkerCount())
				require.EqualValues(t, 4294967296, wc.SizeLimit())

				require.Equal(t, "tmp/1/meta", meta.Path())
				require.Equal(t, fs.FileMode(0o644), meta.BoltDB().Perm())
				require.Equal(t, 200, meta.BoltDB().MaxBatchSize())
				require.Equal(t, 20*time.Millisecond, meta.BoltDB().MaxBatchDelay())

				require.Equal(t, false, sc.Compress())
				require.Equal(t, []string(nil), sc.UncompressableContentTypes())
				require.EqualValues(t, 102400, sc.SmallSizeLimit())

				require.Equal(t, 2, len(ss))

				blz := blobovniczaconfig.From((*config.Config)(ss[0]))
				require.Equal(t, "tmp/1/blob/blobovnicza", ss[0].Path())
				require.EqualValues(t, 4194304, blz.Size())
				require.EqualValues(t, 1, blz.ShallowDepth())
				require.EqualValues(t, 4, blz.ShallowWidth())
				require.EqualValues(t, 50, blz.OpenedCacheSize())
				require.EqualValues(t, 10, blz.LeafWidth())
				require.EqualValues(t, blobovniczaconfig.InitWorkerCountDefault, blz.InitWorkerCount())
				require.EqualValues(t, blobovniczaconfig.RebuildDropTimeoutDefault, blz.RebuildDropTimeout())

				require.Equal(t, "tmp/1/blob", ss[1].Path())
				require.EqualValues(t, 0o644, ss[1].Perm())

				fst := fstreeconfig.From((*config.Config)(ss[1]))
				require.EqualValues(t, 5, fst.Depth())
				require.Equal(t, true, fst.NoSync())

				require.EqualValues(t, 200, gc.RemoverBatchSize())
				require.Equal(t, 5*time.Minute, gc.RemoverSleepInterval())
				require.Equal(t, gcconfig.ExpiredCollectorBatchSizeDefault, gc.ExpiredCollectorBatchSize())
				require.Equal(t, gcconfig.ExpiredCollectorWorkersCountDefault, gc.ExpiredCollectorWorkerCount())

				require.Equal(t, true, sc.RefillMetabase())
				require.Equal(t, mode.ReadWrite, sc.Mode())
			}
			return nil
		})
		require.NoError(t, err)
		require.Equal(t, 2, num)
	}

	configtest.ForEachFileType(path, fileConfigTest)

	t.Run("ENV", func(t *testing.T) {
		configtest.ForEnvFileType(t, path, fileConfigTest)
	})
}
