package blobstor

import (
	"context"
	"encoding/hex"
	"time"

	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/blobstor/common"
	"git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/util/logicerr"
	"git.frostfs.info/TrueCloudLab/frostfs-observability/tracing"
	"git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client"
	apistatus "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/client/status"
	"go.opentelemetry.io/otel/attribute"
	"go.opentelemetry.io/otel/trace"
)

// Get reads the object from b.
// If the descriptor is present, only one sub-storage is tried,
// Otherwise, each sub-storage is tried in order.
func (b *BlobStor) Get(ctx context.Context, prm common.GetPrm) (res common.GetRes, err error) {
	startedAt := time.Now()
	defer func() {
		b.metrics.Get(time.Since(startedAt), len(res.RawData), err == nil, prm.StorageID != nil)
	}()

	ctx, span := tracing.StartSpanFromContext(ctx, "BlobStor.Get",
		trace.WithAttributes(
			attribute.String("address", prm.Address.EncodeToString()),
			attribute.Bool("raw", prm.Raw),
			attribute.String("storage_id", hex.EncodeToString(prm.StorageID)),
		))
	defer span.End()

	b.modeMtx.RLock()
	defer b.modeMtx.RUnlock()

	if prm.StorageID == nil {
		for i := range b.storage {
			res, err = b.storage[i].Storage.Get(ctx, prm)
			if err == nil || !client.IsErrObjectNotFound(err) {
				return res, err
			}
		}

		return common.GetRes{}, logicerr.Wrap(new(apistatus.ObjectNotFound))
	}
	if len(prm.StorageID) == 0 {
		res, err = b.storage[len(b.storage)-1].Storage.Get(ctx, prm)
	} else {
		res, err = b.storage[0].Storage.Get(ctx, prm)
	}
	return res, err
}
